import sys
from typing import Optional, Union
from pydantic import BaseModel, validator

from pogema.utils import CommonSettings

from typing_extensions import Literal


class GridConfig(CommonSettings, ):
    on_target: Literal['finish', 'nothing', 'restart'] = 'finish'
    seed: Optional[int] = None
    size: int = 8
    density: float = 0.3
    num_agents: int = 1
    obs_radius: int = 5
    agents_xy: Optional[list] = None
    targets_xy: Optional[list] = None
    collision_system: Literal['block_both', 'priority', 'soft'] = 'priority'
    persistent: bool = False
    observation_type: Literal['POMAPF', 'MAPF', 'default'] = 'default'
    map: Union[list, str] = None

    map_name: str = None

    integration: Literal['SampleFactory', 'PyMARL', 'rllib', 'gym', 'PettingZoo'] = None
    max_episode_steps: int = 64
    auto_reset: Optional[bool] = None

    @validator('seed')
    def seed_initialization(cls, v):
        assert v is None or (0 <= v < sys.maxsize), "seed must be in [0, " + str(sys.maxsize) + ']'
        return v

    @validator('size')
    def size_restrictions(cls, v):
        assert 2 <= v <= 1024, "size must be in [2, 1024]"
        return v

    @validator('density')
    def density_restrictions(cls, v):
        assert 0.0 <= v <= 1, "density must be in [0, 1]"
        return v

    @validator('num_agents')
    def num_agents_must_be_positive(cls, v):
        assert 1 <= v <= 10000, "num_agents must be in [1, 10000]"
        return v

    @validator('obs_radius')
    def obs_radius_must_be_positive(cls, v):
        assert 1 <= v <= 128, "obs_radius must be in [1, 128]"
        return v

    @validator('map', always=True)
    def map_validation(cls, v, values, ):
        if v is None:
            return None
        if isinstance(v, str):
            v, agents_xy, targets_xy = cls.str_map_to_list(v, values['FREE'], values['OBSTACLE'])
            if agents_xy and targets_xy and values['agents_xy'] is not None and values['targets_xy'] is not None:
                raise KeyError("""Can't create task. Please provide agents_xy and targets_xy only ones.
                Either with parameters or with a map.""")
            elif agents_xy and targets_xy:
                values['agents_xy'] = agents_xy
                values['targets_xy'] = targets_xy
                values['num_agents'] = len(agents_xy)
        size = len(v)
        area = 0
        for line in v:
            size = max(size, len(line))
            area += len(line)
        values['size'] = size
        values['density'] = sum([sum(line) for line in v]) / area
        return v

    @validator('agents_xy')
    def agents_xy_validation(cls, v, values):
        if v is not None:
            cls.check_positions(v, values['size'])
            values['num_agents'] = len(v)
        return v

    @validator('targets_xy')
    def targets_xy_validation(cls, v, values):
        if v is not None:
            cls.check_positions(v, values['size'])
            values['num_agents'] = len(v)
        return v

    @staticmethod
    def check_positions(v, size):
        for position in v:
            x, y = position
            if not (0 <= x < size and 0 <= y < size):
                raise IndexError("Position is out of bounds!")

    @staticmethod
    def str_map_to_list(str_map, free, obstacle):
        obstacles = []
        agents = {}
        targets = {}
        for idx, line in enumerate(str_map.split()):
            row = []
            for char in line:
                if char == '.':
                    row.append(free)
                elif char == '#':
                    row.append(obstacle)
                elif 'A' <= char <= 'Z':
                    targets[char.lower()] = len(obstacles), len(row)
                    row.append(free)
                elif 'a' <= char <= 'z':
                    agents[char.lower()] = len(obstacles), len(row)
                    row.append(free)
                else:
                    raise KeyError(f"Unsupported symbol '{char}' at line {idx}")
            if row:
                if obstacles:
                    assert len(obstacles[-1]) == len(row), f"Wrong string size for row {idx};"
                obstacles.append(row)

        targets_xy = []
        agents_xy = []
        for _, (x, y) in sorted(agents.items()):
            agents_xy.append([x, y])
        for _, (x, y) in sorted(targets.items()):
            targets_xy.append([x, y])

        assert len(targets_xy) == len(agents_xy)
        return obstacles, agents_xy, targets_xy


class PredefinedDifficultyConfig(GridConfig):
    density: float = 0.3
    collision_system: Literal['priority'] = 'priority'
    obs_radius: Literal[5] = 5
    observation_type: Literal['default'] = 'default'

    @validator('density', always=True)
    def density_restrictions(cls, v):
        assert 0.299999 <= v <= 0.3000001, "density for that predefined configuration must be equal to 0.3"
        return v


class Easy8x8(PredefinedDifficultyConfig):
    size: Literal[8] = 8
    max_episode_steps: Literal[64] = 64
    num_agents: Literal[1] = 1
    map_name: Literal['Easy8x8'] = 'Easy8x8'


class Normal8x8(PredefinedDifficultyConfig):
    size: Literal[8] = 8
    max_episode_steps: Literal[64] = 64
    num_agents: Literal[2] = 2
    map_name: Literal['Normal8x8'] = 'Normal8x8'


class Hard8x8(PredefinedDifficultyConfig):
    size: Literal[8] = 8
    max_episode_steps: Literal[64] = 64
    num_agents: Literal[4] = 4
    map_name: Literal['Hard8x8'] = 'Hard8x8'


class ExtraHard8x8(PredefinedDifficultyConfig):
    size: Literal[8] = 8
    max_episode_steps: Literal[64] = 64
    num_agents: Literal[8] = 8
    map_name: Literal['ExtraHard8x8'] = 'ExtraHard8x8'


class Easy16x16(PredefinedDifficultyConfig):
    size: Literal[16] = 16
    max_episode_steps: Literal[128] = 128
    num_agents: Literal[4] = 4
    map_name: Literal['Easy16x16'] = 'Easy16x16'


class Normal16x16(PredefinedDifficultyConfig):
    size: Literal[16] = 16
    max_episode_steps: Literal[128] = 128
    num_agents: Literal[8] = 8
    map_name: Literal['Normal16x16'] = 'Normal16x16'


class Hard16x16(PredefinedDifficultyConfig):
    size: Literal[16] = 16
    max_episode_steps: Literal[128] = 128
    num_agents: Literal[16] = 16
    map_name: Literal['Hard16x16'] = 'Hard16x16'


class ExtraHard16x16(PredefinedDifficultyConfig):
    size: Literal[16] = 16
    max_episode_steps: Literal[128] = 128
    num_agents: Literal[32] = 32
    map_name: Literal['ExtraHard16x16'] = 'ExtraHard16x16'


class Easy32x32(PredefinedDifficultyConfig):
    size: Literal[32] = 32
    max_episode_steps: Literal[256] = 256
    num_agents: Literal[16] = 16
    map_name: Literal['Easy32x32'] = 'Easy32x32'


class Normal32x32(PredefinedDifficultyConfig):
    size: Literal[32] = 32
    max_episode_steps: Literal[256] = 256
    num_agents: Literal[32] = 32
    map_name: Literal['Normal32x32'] = 'Normal32x32'


class Hard32x32(PredefinedDifficultyConfig):
    size: Literal[32] = 32
    max_episode_steps: Literal[256] = 256
    num_agents: Literal[64] = 64
    map_name: Literal['Hard32x32'] = 'Hard32x32'


class ExtraHard32x32(PredefinedDifficultyConfig):
    size: Literal[32] = 32
    max_episode_steps: Literal[256] = 256
    num_agents: Literal[128] = 128
    map_name: Literal['ExtraHard32x32'] = 'ExtraHard32x32'


class Easy64x64(PredefinedDifficultyConfig):
    size: Literal[32] = 64
    max_episode_steps: Literal[512] = 512
    num_agents: Literal[16] = 64
    map_name: Literal['Easy64x64'] = 'Easy64x64'


class Normal64x64(PredefinedDifficultyConfig):
    size: Literal[32] = 64
    max_episode_steps: Literal[512] = 512
    num_agents: Literal[16] = 128
    map_name: Literal['Normal64x64'] = 'Normal64x64'


class Hard64x64(PredefinedDifficultyConfig):
    size: Literal[32] = 64
    max_episode_steps: Literal[512] = 512
    num_agents: Literal[16] = 256
    map_name: Literal['Hard64x64'] = 'Hard64x64'


class ExtraHard64x64(PredefinedDifficultyConfig):
    size: Literal[32] = 64
    max_episode_steps: Literal[512] = 512
    num_agents: Literal[16] = 512
    map_name: Literal['ExtraHard64x64'] = 'ExtraHard64x64'



import numpy as np
from heapq import heappop, heappush
from pogema import GridConfig, pogema_v0
from pogema.animation import AnimationMonitor

INF = 1000000007


class Node:
    def __init__(self, coord=(INF, INF), g: int = 0, h: int = 0):
        self.i, self.j = coord
        self.g = g
        self.h = h
        self.f = g + h

    def __lt__(self, other):
        return self.f < other.f or ((self.f == other.f) and (self.g < other.g))


class AStar:
    def __init__(self):
        self.start = (0, 0)
        self.goal = (0, 0)
        self.max_steps = 500
        self.OPEN = list()
        self.CLOSED = dict()
        self.obstacles = set()
        self.other_agents = set()
        self.best_node = Node(self.start, 0, self.h(self.start))

    def h(self, node):
        return abs(node[0] - self.goal[0]) + abs(node[1] - self.goal[1])

    def get_neighbours(self, u):
        neighbors = []
        for d in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
            if (u[0] + d[0], u[1] + d[1]) not in self.obstacles:
                neighbors.append((u[0] + d[0], u[1] + d[1]))
        return neighbors

    def compute_shortest_path(self):
        u = Node()
        steps = 0
        while len(self.OPEN) > 0 and steps < self.max_steps and (u.i, u.j) != self.goal:
            u = heappop(self.OPEN)
            if self.best_node.h > u.h:
                self.best_node = u
            steps += 1
            for n in self.get_neighbours((u.i, u.j)):
                if n not in self.CLOSED and n not in self.other_agents:
                    heappush(self.OPEN, Node(n, u.g + 1, self.h(n)))
                    self.CLOSED[n] = (u.i, u.j)

    def get_next_node(self):
        next_node = self.start
        if self.goal in self.CLOSED:
            next_node = self.goal
            while self.CLOSED[next_node] != self.start:
                next_node = self.CLOSED[next_node]
        return next_node

    def update_obstacles(self, obs, other_agents):
        obstacles = np.nonzero(obs)
        self.obstacles.clear()
        for k in range(len(obstacles[0])):
            self.obstacles.add((obstacles[0][k], obstacles[1][k]))
        self.other_agents.clear()
        agents = np.nonzero(other_agents)
        for k in range(len(agents[0])):
            self.other_agents.add((agents[0][k], agents[1][k]))

    def reset(self):
        self.CLOSED = dict()
        self.OPEN = list()
        heappush(self.OPEN, Node(self.start, 0, self.h(self.start)))
        self.best_node = Node(self.start, 0, self.h(self.start))

    def update_path(self, start, goal):
        self.start = start
        self.goal = goal
        self.reset()
        self.compute_shortest_path()


class DeterministicPolicy:
    def __init__(self, random_seed=42, random_rate=0.2):
        self.agents = None
        self.actions = {tuple(GridConfig().MOVES[i]): i for i in
                        range(len(GridConfig().MOVES))}  # make a dictionary to translate coordinates of action into id
        self.obs_radius = GridConfig().obs_radius
        self._rnd = np.random.RandomState(random_seed)
        self._random_rate = random_rate

    def get_goal(self, obs):
        goal = np.nonzero(obs[2])
        goal_i = goal[0][0]
        goal_j = goal[1][0]
        if obs[0][goal_i][goal_j]:
            goal_i -= 1 if goal_i == 0 else 0
            goal_i += 1 if goal_i == self.obs_radius * 2 else 0
            goal_j -= 1 if goal_j == 0 else 0
            goal_j += 1 if goal_j == self.obs_radius * 2 else 0
        return goal_i, goal_j

    def act(self, obs) -> list:
        if self.agents is None:
            self.obs_radius = len(obs[0][0]) // 2
            self.agents = [AStar() for _ in range(len(obs))]  # create a planner for each of the agents
        actions = []
        for k in range(len(obs)):
            start = (self.obs_radius, self.obs_radius)
            goal = self.get_goal(obs[k])
            if start == goal:  # don't waste time on the agents that have already reached their goals
                actions.append(0)  # just add useless action to save the order and length of the actions
                continue
            self.agents[k].update_obstacles(obs[k][0], obs[k][1])
            self.agents[k].update_path(start, goal)
            next_node = self.agents[k].get_next_node()
            actions.append(self.actions[(next_node[0] - start[0], next_node[1] - start[1])])
        for idx, action in enumerate(actions):
            if self._rnd.random() < self._random_rate:
                actions[idx] = self._rnd.randint(1, 4)
        return actions


def run_policy(gc: GridConfig, save_animation=False):
    policy = DeterministicPolicy()
    env = pogema_v0(grid_config=gc)
    if save_animation:
        env = AnimationMonitor(env)

    while True:
        obs, info = env.reset()
        while True:
            obs, reward, terminated, truncated, info = env.step(policy.act(obs))
            if all(terminated) or all(truncated):
                break

        yield info[0]['metrics']


def test_life_long():
    gc = GridConfig(num_agents=20, size=8, obs_radius=4, seed=42, max_episode_steps=64, on_target='restart')
    results_generator = run_policy(gc, save_animation=False)

    for _ in range(3):
        metrics = results_generator.__next__()
        assert np.isclose(metrics['avg_throughput'], 1.765625)

    metrics = results_generator.__next__()
    assert np.isclose(metrics['avg_throughput'], 1.71875)
    gc = GridConfig(num_agents=24, size=8, obs_radius=4, seed=43, max_episode_steps=64, on_target='restart')
    results_generator = run_policy(gc, save_animation=False)

    metrics = results_generator.__next__()
    assert np.isclose(metrics['avg_throughput'], 0.453125)


def test_disappearing():
    gc = GridConfig(num_agents=20, size=8, obs_radius=2, seed=42, density=0.2, max_episode_steps=32, on_target='finish')
    results_generator = run_policy(gc, save_animation=False)

    metrics = results_generator.__next__()
    assert np.isclose(metrics['ep_length'], 22.95)
    assert np.isclose(metrics['ISR'], 0.5)
    assert np.isclose(metrics['CSR'], 0.0)

    metrics = results_generator.__next__()
    assert np.isclose(metrics['ep_length'], 15.55)
    assert np.isclose(metrics['ISR'], 0.9)
    assert np.isclose(metrics['CSR'], 0.0)


def test_non_disappearing():
    gc = GridConfig(num_agents=4, size=5, obs_radius=2, seed=3, density=0.2, max_episode_steps=32, on_target='nothing')
    results_generator = run_policy(gc, save_animation=False)

    metrics = results_generator.__next__()
    assert np.isclose(metrics['ep_length'], 20)
    assert np.isclose(metrics['CSR'], 1.0)
    assert np.isclose(metrics['ISR'], 1.0)

    metrics = results_generator.__next__()
    assert np.isclose(metrics['ep_length'], 13)
    assert np.isclose(metrics['CSR'], 1.0)
    assert np.isclose(metrics['ISR'], 1.0)

    gc = GridConfig(num_agents=7, size=5, obs_radius=2, seed=3, density=0.2, max_episode_steps=32, on_target='nothing')
    results_generator = run_policy(gc, save_animation=False)

    metrics = results_generator.__next__()
    assert np.isclose(metrics['ep_length'], 31)
    assert np.isclose(metrics['CSR'], 0.0)
    assert np.isclose(metrics['ISR'], 0.71428571428)



import functools

import numpy as np
from pogema import GridConfig
from pogema.envs import _make_pogema


def parallel_env(grid_config: GridConfig = GridConfig()):
    return PogemaParallel(grid_config)


class PogemaParallel:

    def state(self):
        return self.pogema.get_state()

    def __init__(self, grid_config: GridConfig, render_mode='ansi'):
        self.metadata = {'render_modes': ['ansi'], "name": "pogema"}
        self.render_mode = render_mode
        self.pogema = _make_pogema(grid_config)
        self.possible_agents = ["player_" + str(r) for r in range(self.pogema.get_num_agents())]
        self.agent_name_mapping = dict(zip(self.possible_agents, list(range(len(self.possible_agents)))))
        self.agents = None
        self.num_moves = None

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent):
        assert agent in self.possible_agents
        return self.pogema.observation_space

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent):
        assert agent in self.possible_agents
        return self.pogema.action_space

    def render(self, mode="human"):
        assert mode == 'human'
        return self.pogema.render()

    def reset(self, seed=None, options=None):
        observations, info = self.pogema.reset(seed=seed, options=options)
        self.agents = self.possible_agents[:]
        self.num_moves = 0
        observations = {agent: observations[self.agent_name_mapping[agent]].astype(np.float32) for agent in self.agents}
        return observations

    def step(self, actions):
        anm = self.agent_name_mapping

        actions = [actions[agent] if agent in actions else 0 for agent in self.possible_agents]
        observations, rewards, terminated, truncated, infos = self.pogema.step(actions)
        d_observations = {agent: observations[anm[agent]].astype(np.float32) for agent in
                          self.agents}
        d_rewards = {agent: rewards[anm[agent]] for agent in self.agents}
        d_terminated = {agent: terminated[anm[agent]] for agent in self.agents}
        d_truncated = {agent: truncated[anm[agent]] for agent in self.agents}
        d_infos = {agent: infos[anm[agent]] for agent in self.agents}

        for agent, idx in anm.items():
            if (not self.pogema.grid.is_active[idx] or all(truncated) or all(terminated)) and agent in self.agents:
                self.agents.remove(agent)

        return d_observations, d_rewards, d_terminated, d_truncated, d_infos

    @property
    def unwrapped(self):
        return self

    def close(self):
        pass



import numpy as np

from pogema.utils import check_grid, render_grid

GRID_STR_REGISTRY = {}


def in_registry(name):
    return name in GRID_STR_REGISTRY


def get_grid(name):
    if in_registry(name):
        return GRID_STR_REGISTRY[name]
    else:
        raise KeyError(f"Grid with name {name} not found")


class RegisteredGrid:
    FREE = 0
    OBSTACLE = 1

    def str_to_grid(self, grid_str):
        obstacles = []
        agents = {}
        targets = {}
        for idx, line in enumerate(grid_str.split()):
            row = []
            for char in line:
                if char == '.':
                    row.append(self.FREE)
                elif char == '#':
                    row.append(self.OBSTACLE)
                elif 'A' <= char <= 'Z':
                    targets[char.lower()] = len(obstacles), len(row)
                    row.append(self.FREE)
                elif 'a' <= char <= 'z':
                    agents[char.lower()] = len(obstacles), len(row)
                    row.append(self.FREE)
                else:
                    raise KeyError(f"Unsupported symbol '{char}' at line {idx}")
            if row:
                if obstacles:
                    assert len(obstacles[-1]) == len(row), f"Wrong string size for row {idx};"
                obstacles.append(row)
        return obstacles, agents, targets

    def __init__(self, name: str, grid_str: str = None, agents_positions: list = None, agents_targets: list = None):
        self.name = name
        self.grid_str = grid_str
        self.agents_positions = agents_positions
        self.agents_targets = agents_targets

        self.obstacles, agents, targets = self.str_to_grid(grid_str)
        self.obstacles = np.array(self.obstacles, dtype=np.int32)

        if agents_positions and agents:
            raise ValueError("Agents positions are already defined in the grid string!")
        if agents_targets and targets:
            raise ValueError("Agents targets are already defined in the grid string!")

        if agents:
            self.agents_xy = []
            for _, (x, y) in sorted(agents.items()):
                self.agents_xy.append([x, y])
        else:
            self.agents_xy = agents_positions

        if targets:
            self.targets_xy = []
            for _, (x, y) in sorted(targets.items()):
                self.targets_xy.append([x, y])
        else:
            self.targets_xy = agents_targets
        global GRID_STR_REGISTRY
        if in_registry(name):
            raise ValueError(f"Grid with name {self.name} already registered!")
        check_grid(self.obstacles, self.agents_xy, self.targets_xy)

        register_grid(self)

    def get_obstacles(self):
        return self.obstacles

    def get_agents_xy(self):
        return self.agents_xy

    def get_targets_xy(self):
        return self.targets_xy

    def render(self):
        render_grid(obstacles=self.get_obstacles(), positions_xy=self.get_agents_xy(), targets_xy=self.get_targets_xy())


def register_grid(rg: RegisteredGrid):
    if in_registry(rg.name):
        raise KeyError(f"Grid with name {rg.name} already registered")
    GRID_STR_REGISTRY[rg.name] = rg



from gymnasium import Wrapper


class AbstractMetric(Wrapper):
    def _compute_stats(self, step, is_on_goal, finished):
        raise NotImplementedError

    def __init__(self, env):
        super().__init__(env)
        self._current_step = 0

    def step(self, action):
        obs, reward, terminated, truncated, infos = self.env.step(action)
        finished = all(truncated) or all(terminated)

        metric = self._compute_stats(self._current_step, self.was_on_goal, finished)
        self._current_step += 1
        if finished:
            self._current_step = 0

        if metric:
            if 'metrics' not in infos[0]:
                infos[0]['metrics'] = {}
            infos[0]['metrics'].update(**metric)

        return obs, reward, terminated, truncated, infos


class LifeLongAverageThroughputMetric(AbstractMetric):

    def __init__(self, env):
        super().__init__(env)
        self._solved_instances = 0

    def _compute_stats(self, step, is_on_goal, finished):
        for agent_idx, on_goal in enumerate(is_on_goal):
            if on_goal:
                self._solved_instances += 1
        if finished:
            result = {'avg_throughput': self._solved_instances / self.grid_config.max_episode_steps}
            self._solved_instances = 0
            return result


class LifeLongAttritionMetric(AbstractMetric):

    def __init__(self, env):
        super().__init__(env)
        self._attrition_steps = 0
        self._on_goal_steps = 0

    def _compute_stats(self, step, is_on_goal, finished):
        for agent_idx, on_goal in enumerate(is_on_goal):
            if not on_goal:
                self._attrition_steps += 1
            else:
                self._on_goal_steps += 1
        if finished:
            result = {
                'attrition': self._attrition_steps / max(1, self._on_goal_steps)}
            self._solved_instances = 0
            return result


class NonDisappearCSRMetric(AbstractMetric):

    def _compute_stats(self, step, is_on_goal, finished):
        if finished:
            return {'CSR': float(all(is_on_goal))}


class NonDisappearISRMetric(AbstractMetric):

    def _compute_stats(self, step, is_on_goal, finished):
        if finished:
            return {'ISR': float(sum(is_on_goal)) / self.get_num_agents()}


class NonDisappearEpLengthMetric(AbstractMetric):

    def _compute_stats(self, step, is_on_goal, finished):
        if finished:
            return {'ep_length': step}


class EpLengthMetric(AbstractMetric):
    def __init__(self, env):
        super().__init__(env)
        self._solve_time = [None for _ in range(self.get_num_agents())]

    def _compute_stats(self, step, is_on_goal, finished):
        for idx, on_goal in enumerate(is_on_goal):
            if self._solve_time[idx] is None:
                if on_goal or finished:
                    self._solve_time[idx] = step

        if finished:
            result = {'ep_length': sum(self._solve_time) / self.get_num_agents() + 1}
            self._solve_time = [None for _ in range(self.get_num_agents())]
            return result


class CSRMetric(AbstractMetric):
    def __init__(self, env):
        super().__init__(env)
        self._solved_instances = 0

    def _compute_stats(self, step, is_on_goal, finished):
        self._solved_instances += sum(is_on_goal)
        if finished:
            results = {'CSR': float(self._solved_instances == self.get_num_agents())}
            self._solved_instances = 0
            return results


class ISRMetric(AbstractMetric):
    def __init__(self, env):
        super().__init__(env)
        self._solved_instances = 0

    def _compute_stats(self, step, is_on_goal, finished):
        self._solved_instances += sum(is_on_goal)
        if finished:
            results = {'ISR': self._solved_instances / self.get_num_agents()}
            self._solved_instances = 0
            return results



import re
import time

import numpy as np
from tabulate import tabulate

from pogema import pogema_v0
from pogema import Easy8x8, Normal8x8, Hard8x8, ExtraHard8x8
from pogema import Easy16x16, Normal16x16, Hard16x16, ExtraHard16x16
from pogema import Easy32x32, Normal32x32, Hard32x32, ExtraHard32x32
from pogema import Easy64x64, Normal64x64, Hard64x64, ExtraHard64x64

from pogema.animation import AnimationMonitor
from pogema.envs import ActionsSampler
from pogema.grid import GridConfig


class ActionMapping:
    noop: int = 0
    up: int = 1
    down: int = 2
    left: int = 3
    right: int = 4


def test_moving():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42))
    ac = ActionMapping()
    env.reset()

    env.step([ac.right, ac.noop])
    env.step([ac.up, ac.noop])
    env.step([ac.left, ac.noop])
    env.step([ac.down, ac.noop])
    env.step([ac.down, ac.noop])
    env.step([ac.left, ac.noop])
    env.step([ac.left, ac.noop])
    env.step([ac.up, ac.noop])
    env.step([ac.up, ac.noop])
    env.step([ac.up, ac.noop])

    env.step([ac.right, ac.noop])
    env.step([ac.up, ac.noop])
    env.step([ac.right, ac.noop])
    env.step([ac.down, ac.noop])
    obs, reward, terminated, truncated, infos = env.step([ac.right, ac.noop])

    assert np.isclose([1.0, 0.0], reward).all()
    assert np.isclose([True, False], terminated).all()


def test_types():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42))
    obs, info = env.reset()
    assert obs[0].dtype == np.float32


def run_episode(grid_config=None, env=None):
    if env is None:
        env = pogema_v0(grid_config)
    env.reset()

    obs, rewards, terminated, truncated, infos = env.reset(), [None], [False], [False], [None]

    results = [[obs, rewards, terminated, truncated, infos]]
    while True:
        results.append(env.step(env.sample_actions()))
        terminated, truncated = results[-1][2], results[-1][3]
        if all(terminated) or all(truncated):
            break
    return results


def test_metrics():
    *_, infos = run_episode(GridConfig(num_agents=2, seed=5, size=5, max_episode_steps=64))[-1]
    assert np.isclose(infos[0]['metrics']['CSR'], 0.0)
    assert np.isclose(infos[0]['metrics']['ISR'], 0.5)

    *_, infos = run_episode(GridConfig(num_agents=2, seed=5, size=5, max_episode_steps=512))[-1]
    assert np.isclose(infos[0]['metrics']['CSR'], 1.0)
    assert np.isclose(infos[0]['metrics']['ISR'], 1.0)

    *_, infos = run_episode(GridConfig(num_agents=5, seed=5, size=5, max_episode_steps=64))[-1]
    assert np.isclose(infos[0]['metrics']['CSR'], 0.0)
    assert np.isclose(infos[0]['metrics']['ISR'], 0.2)


def test_standard_pogema():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42, on_target='finish'))
    env.reset()
    run_episode(env=env)


def test_pomapf_observation():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42, on_target='finish',
                               observation_type='POMAPF'))
    obs, info = env.reset()
    assert 'agents' in obs[0]
    assert 'obstacles' in obs[0]
    assert 'xy' in obs[0]
    assert 'target_xy' in obs[0]
    run_episode(env=env)


def test_mapf_observation():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42, on_target='finish',
                               observation_type='MAPF'))
    obs, info = env.reset()
    assert 'global_obstacles' in obs[0]
    assert 'global_xy' in obs[0]
    assert 'global_target_xy' in obs[0]
    run_episode(env=env)


def test_standard_pogema_animation():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42, on_target='finish'))
    env = AnimationMonitor(env)
    env.reset()
    run_episode(env=env)


def test_gym_pogema_animation():
    import gymnasium
    env = gymnasium.make('Pogema-v0',
                   grid_config=GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42, on_target='finish'))
    env = AnimationMonitor(env)
    env.reset()
    done = False
    while True:
        _, _, terminated, truncated, _ = env.step(env.action_space.sample())
        if terminated or truncated:
            break

def test_non_disappearing_pogema():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42, on_target='nothing'))
    env.reset()
    run_episode(env=env)


def test_non_disappearing_pogema_no_seed():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=None, on_target='nothing'))
    env.reset()
    run_episode(env=env)


def test_non_disappearing_pogema_animation():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42, on_target='nothing'))
    env = AnimationMonitor(env)
    env.reset()
    run_episode(env=env)


def test_life_long_pogema():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42, on_target='restart'))
    env.reset()
    run_episode(env=env)


def test_life_long_pogema_empty_seed():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=None, on_target='restart'))
    env.reset()
    run_episode(env=env)


def test_life_long_pogema_animation():
    env = pogema_v0(GridConfig(num_agents=2, size=6, obs_radius=2, density=0.3, seed=42, on_target='restart'))
    env = AnimationMonitor(env)
    env.reset()
    run_episode(env=env)


def test_predefined_configurations():
    def get_num_agents_by_target_density(size, agent_density, obstacle_density):
        return round(agent_density * (size * size * (1.0 - obstacle_density)))

    def get_target_density_by_num_agents(size, num_agents, obstacle_density):
        return num_agents / (size * size * (1.0 - obstacle_density))

    predefined_grids = [
        Easy8x8, Normal8x8, Hard8x8, ExtraHard8x8,
        Easy16x16, Normal16x16, Hard16x16, ExtraHard16x16,
        Easy32x32, Normal32x32, Hard32x32, ExtraHard32x32,
        Easy64x64, Normal64x64, Hard64x64, ExtraHard64x64,
    ]

    # checking that the number of agents (agent density) is correct
    for make_grid_config_func in predefined_grids:
        gc = make_grid_config_func(seed=42)
        for difficulty, agent_density in zip(['Easy', 'Normal', 'Hard', 'ExtraHard'],
                                             [0.02232142, 0.04464285, 0.08928571, 0.17857142]):
            if re.match(rf'^{difficulty}\d+x\d+', make_grid_config_func.__name__):
                assert np.isclose(get_target_density_by_num_agents(gc.size, gc.num_agents, gc.density), agent_density)

    # checking creation
    for make_grid_config_func in predefined_grids:
        gc = make_grid_config_func(seed=42)
        env = pogema_v0(gc)
        env.reset()

    # checking map_name
    for make_grid_config_func in predefined_grids:
        gc = make_grid_config_func(seed=42)
        assert gc.map_name == make_grid_config_func.__name__


def test_persistent_env(num_steps=100):
    seed = 42

    env = pogema_v0(
        grid_config=GridConfig(on_target='finish', seed=seed, num_agents=8, density=0.132, size=8, obs_radius=2,
                               persistent=True))

    env.reset()
    action_sampler = ActionsSampler(env.action_space.n, seed=seed)

    first_run_observations = []

    def state_repr(observations, rewards, terminates, truncates, infos):
        return np.concatenate([np.array(observations).flatten(), terminates, truncates, np.array(rewards), ])

    for current_step in range(num_steps):
        actions = action_sampler.sample_actions(dim=env.get_num_agents())
        obs, reward, terminated, truncated, info = env.step(actions)

        first_run_observations.append(state_repr(obs, reward, terminated, truncated, info))
        if all(terminated) or all(truncated):
            break

    # resetting the environment to the initial state using backward steps
    for current_step in range(num_steps):
        if not env.step_back():
            break

    action_sampler = ActionsSampler(env.action_space.n, seed=seed)

    second_run_observations = []
    for current_step in range(num_steps):
        actions = action_sampler.sample_actions(dim=env.get_num_agents())
        obs, reward, terminated, truncated, info = env.step(actions)
        second_run_observations.append(state_repr(obs, reward, terminated, truncated, info))
        assert np.isclose(first_run_observations[current_step], second_run_observations[current_step]).all()
        if all(terminated) or all(truncated):
            break
    assert np.isclose(first_run_observations, second_run_observations).all()


def test_steps_per_second_throughput():
    table = []
    for on_target in ['finish', 'nothing', 'restart']:
        for num_agents in [1, 32, 64]:
            for size in [32, 64]:
                gc = GridConfig(obs_radius=5, seed=42, max_episode_steps=1024, )
                gc.size = size
                gc.num_agents = num_agents
                gc.on_target = on_target

                start_time = time.monotonic()
                run_episode(grid_config=gc)
                end_time = time.monotonic()
                steps_per_second = gc.max_episode_steps / (end_time - start_time)
                table.append([on_target, num_agents, size, steps_per_second * gc.num_agents])
    print('\n' + tabulate(table, headers=['on_target', 'num_agents', 'size', 'SPS (individual)'], tablefmt='grid'))



import numpy as np

from pogema import GridConfig
from pogema.envs import _make_pogema


class PyMarlPogema:

    def __init__(self, grid_config, mh_distance=False):
        gc = grid_config
        self._grid_config: GridConfig = gc

        self.env = _make_pogema(grid_config)
        self._mh_distance = mh_distance
        self._observations, _ = self.env.reset()
        self.max_episode_steps = gc.max_episode_steps
        self.episode_limit = gc.max_episode_steps
        self.n_agents = self.env.get_num_agents()

        self.spec = None

    @property
    def unwrapped(self):
        return self

    def step(self, actions):
        self._observations, rewards, terminated, truncated, infos = self.env.step(actions)
        info = {}
        done = all(terminated) or all(truncated)
        if done:
            for key, value in infos[0]['metrics'].items():
                info[key] = value
            
        return sum(rewards), done, info

    def get_obs(self):
        return np.array([self.get_obs_agent(agent_id) for agent_id in range(self.n_agents)])

    def get_obs_agent(self, agent_id):
        return np.array(self._observations[agent_id]).flatten()

    def get_obs_size(self):
        return len(np.array(self._observations[0]).flatten())

    def get_state(self):
        return self.env.get_state()

    def get_state_size(self):
        return len(self.get_state())

    def get_avail_actions(self):
        actions = []
        for i in range(self.env.get_num_agents()):
            actions.append(self.get_avail_agent_actions(i))
        return actions

    # noinspection PyUnusedLocal
    @staticmethod
    def get_avail_agent_actions(agent_id):
        return list(range(5))

    @staticmethod
    def get_total_actions():
        return 5

    def reset(self):
        self._grid_config = self.env.grid_config
        self._observations, _ = self.env.reset()
        return np.array(self._observations).flatten()

    def save_replay(self):
        return

    def render(self, *args, **kwargs):
        return self.env.render(*args, **kwargs)

    def get_env_info(self):
        env_info = {"state_shape": self.get_state_size(),
                    "obs_shape": self.get_obs_size(),
                    "n_actions": self.get_total_actions(),
                    "n_agents": self.n_agents,
                    "episode_limit": self.episode_limit,
                    }
        return env_info

    @staticmethod
    def get_stats():
        return {}

    def close(self):
        return

    def sample_actions(self):
        return self.env.sample_actions()



from gymnasium import register
from pogema.grid_config import GridConfig
from pogema.integrations.make_pogema import pogema_v0
from pogema.a_star_policy import AStarAgent, BatchAStarAgent

from pogema.grid_config import Easy8x8, Normal8x8, Hard8x8, ExtraHard8x8
from pogema.grid_config import Easy16x16, Normal16x16, Hard16x16, ExtraHard16x16
from pogema.grid_config import Easy32x32, Normal32x32, Hard32x32, ExtraHard32x32
from pogema.grid_config import Easy64x64, Normal64x64, Hard64x64, ExtraHard64x64

__version__ = '1.2.3a0'

__all__ = [
    'GridConfig',
    'pogema_v0',
    'AStarAgent', 'BatchAStarAgent',
    'Easy8x8', 'Normal8x8', 'Hard8x8', 'ExtraHard8x8',
    'Easy16x16', 'Normal16x16', 'Hard16x16', 'ExtraHard16x16',
    'Easy32x32', 'Normal32x32', 'Hard32x32', 'ExtraHard32x32',
    'Easy64x64', 'Normal64x64', 'Hard64x64', 'ExtraHard64x64',
]

register(
    id="Pogema-v0",
    entry_point="pogema.integrations.make_pogema:make_single_agent_gym",
)



import math
import os
import typing
from dataclasses import dataclass
from itertools import cycle
from gymnasium import logger, Wrapper

from pogema import GridConfig, pogema_v0, BatchAStarAgent
from pogema.svg_objects import Line, RectangleHref, Animation, Circle, Rectangle
from pogema.grid import Grid
from pogema.wrappers.persistence import PersistentWrapper, AgentState


@dataclass
class AnimationSettings:
    """
    Settings for the animation.
    """
    r: int = 35
    stroke_width: int = 10
    scale_size: int = 100
    time_scale: float = 0.25
    draw_start: int = 100
    rx: int = 15

    obstacle_color: str = '#84A1AE'
    ego_color: str = '#c1433c'
    ego_other_color: str = '#6e81af'
    shaded_opacity: float = 0.2
    egocentric_shaded: bool = True
    stroke_dasharray: int = 25

    colors: tuple = (
        '#c1433c',
        '#2e6f9e',
        '#6e81af',
        '#00b9c8',
        '#72D5C8',
        '#0ea08c',
        '#8F7B66',
    )


@dataclass
class AnimationConfig:
    """
    Configuration for the animation.
    """
    directory: str = 'renders/'
    static: bool = False
    show_agents: bool = True
    egocentric_idx: typing.Optional[int] = None
    uid: typing.Optional[str] = None
    save_every_idx_episode: typing.Optional[int] = 1
    show_lines: bool = True


@dataclass
class GridHolder:
    """
    Holds the grid and the history.
    """
    obstacles: typing.Any = None
    episode_length: int = None
    height: int = None
    width: int = None
    colors: dict = None
    history: list = None


class Drawing:
    """
    Drawing, analog of the DrawSvg class in the pogema package.
    """

    def __init__(self, height, width, svg_settings, display_inline=False, origin=(0, 0)):
        self.height = height
        self.width = width
        self.display_inline = display_inline
        self.origin = origin
        self.elements = []
        self.svg_settings = svg_settings

    def add_element(self, element):
        self.elements.append(element)

    def render(self):
        view_box = (0, -self.height, self.width, self.height)
        width = self.width
        height = self.height
        t = max(height, width) / 512
        width = math.ceil(width / t)
        height = math.ceil(height / t)

        results = [f'''<?xml version="1.0" encoding="UTF-8"?>
        <svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"
             width="{width}" height="{height}" viewBox="{" ".join(map(str, view_box))}">''',
                   '\n<defs>\n',
                   f'<rect id="obstacle" width="{self.svg_settings.r * 2}" height="{self.svg_settings.r * 2}" fill="{self.svg_settings.obstacle_color}" rx="{self.svg_settings.rx}"/>',
                   '',
                   '<style>',
                   '.line {' + f'stroke: {self.svg_settings.obstacle_color}; stroke-width: {self.svg_settings.stroke_width};' + '}',
                   '.agent {' + f'r: {self.svg_settings.r}' + '}',
                   '.target {' + f'fill: none; stroke-width: {self.svg_settings.stroke_width}; r: {self.svg_settings.r}' + '}',
                   # '.rect { fill: #84A1AE; rx: 15;}',
                   '</style>',
                   '</defs>\n']
        for element in self.elements:
            results.append(element.render())
        results.append('</svg>')
        return "\n".join(results)


class AnimationMonitor(Wrapper):
    """
    Defines the animation, which saves the episode as SVG.
    """

    def __init__(self, env, animation_config=AnimationConfig()):
        # Wrapping env using PersistenceWrapper for saving the history.
        env = PersistentWrapper(env)
        super().__init__(env)

        self.history = self.env.get_history()

        self.svg_settings: AnimationSettings = AnimationSettings()
        self.animation_config: AnimationConfig = animation_config

        self._episode_idx = 0

    def step(self, action):
        """
        Saves information about the episode.
        :param action: current actions
        :return: obs, reward, done, info
        """
        obs, reward, terminated, truncated, info = self.env.step(action)

        multi_agent_terminated = isinstance(terminated, (list, tuple)) and all(terminated)
        single_agent_terminated = isinstance(terminated, (bool, int)) and terminated
        multi_agent_truncated = isinstance(truncated, (list, tuple)) and all(truncated)
        single_agent_truncated = isinstance(truncated, (bool, int)) and truncated

        if multi_agent_terminated or single_agent_terminated or multi_agent_truncated or single_agent_truncated:
            save_tau = self.animation_config.save_every_idx_episode
            if save_tau:
                if (self._episode_idx + 1) % save_tau or save_tau == 1:
                    if not os.path.exists(self.animation_config.directory):
                        logger.info(f"Creating pogema monitor directory {self.animation_config.directory}", )
                        os.makedirs(self.animation_config.directory, exist_ok=True)

                    path = os.path.join(self.animation_config.directory,
                                        self.pick_name(self.grid_config, self._episode_idx))
                    self.save_animation(path)

        return obs, reward, terminated, truncated, info

    @staticmethod
    def pick_name(grid_config: GridConfig, episode_idx=None, zfill_ep=5):
        """
        Picks a name for the SVG file.
        :param grid_config: configuration of the grid
        :param episode_idx: idx of the episode
        :param zfill_ep: zfill for the episode number
        :return:
        """
        gc = grid_config
        name = 'pogema'
        if episode_idx is not None:
            name += f'-ep{str(episode_idx).zfill(zfill_ep)}'
        if gc:
            if gc.map_name:
                name += f'-{gc.map_name}'
            if gc.seed is not None:
                name += f'-seed{gc.seed}'
        else:
            name += '-render'
        return name + '.svg'

    def reset(self, **kwargs):
        """
        Resets the environment and resets the current positions of agents and targets
        :param kwargs:
        :return: obs: observation
        """
        obs = self.env.reset(**kwargs)

        self._episode_idx += 1
        self.history = self.env.get_history()

        return obs

    def create_animation(self, animation_config=None):
        """
        Creates the animation.
        :param animation_config: configuration of the animation
        :return: drawing: drawing object
        """
        anim_cfg = animation_config
        if anim_cfg is None:
            anim_cfg = self.animation_config

        grid: Grid = self.grid
        cfg = self.svg_settings
        colors = cycle(cfg.colors)
        agents_colors = {index: next(colors) for index in range(self.grid_config.num_agents)}

        if anim_cfg.egocentric_idx is not None:
            anim_cfg.egocentric_idx %= self.grid_config.num_agents

        decompressed_history: list[list[AgentState]] = self.env.decompress_history(self.history)

        # Change episode length for egocentric environment
        if anim_cfg.egocentric_idx is not None:
            episode_length = decompressed_history[anim_cfg.egocentric_idx][-1].step + 1
            for agent_idx in range(self.grid_config.num_agents):
                decompressed_history[agent_idx] = decompressed_history[agent_idx][:episode_length]
        else:
            episode_length = len(decompressed_history[0])

        # Add last observation one more time to highlight the final state
        for agent_idx in range(self.grid_config.num_agents):
            decompressed_history[agent_idx].append(decompressed_history[agent_idx][-1])

        # Change episode length for static environment
        if anim_cfg.static:
            episode_length = 1
            decompressed_history = [[decompressed_history[idx][-1]] for idx in range(len(decompressed_history))]

        gh = GridHolder(width=len(grid.obstacles), height=len(grid.obstacles[0]),
                        obstacles=grid.obstacles,
                        colors=agents_colors,
                        episode_length=episode_length,
                        history=decompressed_history, )

        render_width, render_height = gh.height * cfg.scale_size + cfg.scale_size, gh.width * cfg.scale_size + cfg.scale_size

        drawing = Drawing(width=render_width, height=render_height, svg_settings=self.svg_settings)
        obstacles = self.create_obstacles(gh, anim_cfg)

        agents = []
        targets = []

        if anim_cfg.show_agents:
            agents = self.create_agents(gh, anim_cfg)
            targets = self.create_targets(gh, anim_cfg)

            if not anim_cfg.static:
                self.animate_agents(agents, anim_cfg.egocentric_idx, gh)
                self.animate_targets(targets, gh, anim_cfg)
        if anim_cfg.show_lines:
            grid_lines = self.create_grid_lines(gh, render_width, render_height)
            for line in grid_lines:
                drawing.add_element(line)
        for obj in [*obstacles, *agents, *targets, ]:
            drawing.add_element(obj)

        if anim_cfg.egocentric_idx is not None:
            field_of_view = self.create_field_of_view(grid_holder=gh, animation_config=anim_cfg)
            if not anim_cfg.static:
                self.animate_obstacles(obstacles=obstacles, grid_holder=gh, animation_config=anim_cfg)
                self.animate_field_of_view(field_of_view, anim_cfg.egocentric_idx, gh)
            drawing.add_element(field_of_view)

        return drawing

    def create_grid_lines(self, grid_holder: GridHolder, render_width, render_height):
        cfg, r = self.svg_settings, self.grid_config.obs_radius
        offset = (r - 1) * cfg.scale_size
        stroke_settings = {'class': 'line'}
        grid_lines = []

        for i in range(-1 + r, grid_holder.height + 2 - r):
            x = i * cfg.scale_size + cfg.scale_size / 2
            grid_lines.append(Line(x1=x, y1=offset, x2=x, y2=render_height - offset, **stroke_settings))

        for i in range(-1 + r, grid_holder.width + 2 - r):
            y = i * cfg.scale_size + cfg.scale_size / 2
            grid_lines.append(Line(x1=offset, y1=y, x2=render_width - offset, y2=y, **stroke_settings))

        return grid_lines

    def save_animation(self, name='render.svg', animation_config: typing.Optional[AnimationConfig] = None):
        """
        Saves the animation.
        :param name: name of the file
        :param animation_config: animation configuration
        :return: None
        """
        animation = self.create_animation(animation_config)
        with open(name, "w") as f:
            f.write(animation.render())

    @staticmethod
    def fix_point(x, y, length):
        return length - y - 1, x

    @staticmethod
    def check_in_radius(x1, y1, x2, y2, r) -> bool:
        return x2 - r <= x1 <= x2 + r and y2 - r <= y1 <= y2 + r

    def create_field_of_view(self, grid_holder, animation_config):
        """
        Creates the field of view for the egocentric agent.
        :param grid_holder:
        :param animation_config:
        :return:
        """
        cfg = self.svg_settings
        gh: GridHolder = grid_holder
        ego_idx = animation_config.egocentric_idx
        x, y = gh.history[ego_idx][0].get_xy()
        cx = cfg.draw_start + y * cfg.scale_size
        cy = cfg.draw_start + (gh.width - x - 1) * cfg.scale_size

        dr = (self.grid_config.obs_radius + 1) * cfg.scale_size - cfg.stroke_width * 2
        result = Rectangle(x=cx - dr + cfg.r, y=cy - dr + cfg.r,
                           width=2 * dr - 2 * cfg.r, height=2 * dr - 2 * cfg.r,
                           stroke=cfg.ego_color, stroke_width=cfg.stroke_width,
                           fill='none',
                           rx=cfg.rx, stroke_dasharray=cfg.stroke_dasharray,
                           )

        return result

    def animate_field_of_view(self, view, agent_idx, grid_holder):
        """
        Animates the field of view.
        :param view:
        :param agent_idx:
        :param grid_holder:
        :return:
        """
        gh: GridHolder = grid_holder
        cfg = self.svg_settings
        x_path = []
        y_path = []
        for state in gh.history[agent_idx]:
            x, y = state.get_xy()
            dr = (self.grid_config.obs_radius + 1) * cfg.scale_size - cfg.stroke_width * 2
            cx = cfg.draw_start + y * cfg.scale_size
            cy = -cfg.draw_start + -(gh.width - x - 1) * cfg.scale_size
            x_path.append(str(cx - dr + cfg.r))
            y_path.append(str(cy - dr + cfg.r))

        visibility = ['visible' if state.is_active() else 'hidden' for state in gh.history[agent_idx]]

        view.add_animation(self.compressed_anim('x', x_path, cfg.time_scale))
        view.add_animation(self.compressed_anim('y', y_path, cfg.time_scale))
        view.add_animation(self.compressed_anim('visibility', visibility, cfg.time_scale))

    def animate_agents(self, agents, egocentric_idx, grid_holder):
        """
        Animates the agents.
        :param agents:
        :param egocentric_idx:
        :param grid_holder:
        :return:
        """
        gh: GridHolder = grid_holder
        cfg = self.svg_settings
        for agent_idx, agent in enumerate(agents):
            x_path = []
            y_path = []
            opacity = []
            for agent_state in gh.history[agent_idx]:
                x, y = agent_state.get_xy()

                x_path.append(str(cfg.draw_start + y * cfg.scale_size))
                y_path.append(str(-cfg.draw_start + -(gh.width - x - 1) * cfg.scale_size))

                if egocentric_idx is not None:
                    ego_x, ego_y = gh.history[egocentric_idx][idx].get_xy()
                    if self.check_in_radius(x, y, ego_x, ego_y, self.grid_config.obs_radius):
                        opacity.append('1.0')
                    else:
                        opacity.append(str(cfg.shaded_opacity))

            visibility = ['visible' if state.is_active() else 'hidden' for state in gh.history[agent_idx]]

            agent.add_animation(self.compressed_anim('cy', y_path, cfg.time_scale))
            agent.add_animation(self.compressed_anim('cx', x_path, cfg.time_scale))
            agent.add_animation(self.compressed_anim('visibility', visibility, cfg.time_scale))
            if opacity:
                agent.add_animation(self.compressed_anim('opacity', opacity, cfg.time_scale))

    @classmethod
    def compressed_anim(cls, attr_name, tokens, time_scale, rep_cnt='indefinite'):
        """
        Compresses the animation.
        :param attr_name:
        :param tokens:
        :param time_scale:
        :param rep_cnt:
        :return:
        """
        tokens, times = cls.compress_tokens(tokens)
        cumulative = [0, ]
        for t in times:
            cumulative.append(cumulative[-1] + t)
        times = [str(round(value / cumulative[-1], 10)) for value in cumulative]
        tokens = [tokens[0]] + tokens

        times = times
        tokens = tokens
        return Animation(attributeName=attr_name,
                         dur=f'{time_scale * (-1 + cumulative[-1])}s',
                         values=";".join(tokens),
                         repeatCount=rep_cnt,
                         keyTimes=";".join(times))

    @staticmethod
    def wisely_add(token, cnt, tokens, times):
        """
        Adds the token to the tokens and times.
        :param token:
        :param cnt:
        :param tokens:
        :param times:
        :return:
        """
        if cnt > 1:
            tokens += [token, token]
            times += [1, cnt - 1]
        else:
            tokens.append(token)
            times.append(cnt)

    @classmethod
    def compress_tokens(cls, input_tokens: list):
        """
        Compresses the tokens.
        :param input_tokens:
        :return:
        """
        tokens = []
        times = []
        if input_tokens:
            cur_idx = 0
            cnt = 1
            for idx in range(1, len(input_tokens)):
                if input_tokens[idx] == input_tokens[cur_idx]:
                    cnt += 1
                else:
                    cls.wisely_add(input_tokens[cur_idx], cnt, tokens, times)
                    cnt = 1
                    cur_idx = idx
            cls.wisely_add(input_tokens[cur_idx], cnt, tokens, times)
        return tokens, times

    def animate_targets(self, targets, grid_holder, animation_config):
        """
        Animates the targets.
        :param targets:
        :param grid_holder:
        :param animation_config:
        :return:
        """
        gh: GridHolder = grid_holder
        cfg = self.svg_settings
        ego_idx = animation_config.egocentric_idx

        for agent_idx, target in enumerate(targets):
            target_idx = ego_idx if ego_idx is not None else agent_idx

            x_path = []
            y_path = []

            for step_idx, state in enumerate(gh.history[target_idx]):
                x, y = state.get_target_xy()
                x_path.append(str(cfg.draw_start + y * cfg.scale_size))
                y_path.append(str(-cfg.draw_start + -(gh.width - x - 1) * cfg.scale_size))

            visibility = ['visible' if state.is_active() else 'hidden' for state in gh.history[agent_idx]]

            if self.grid_config.on_target == 'restart':
                target.add_animation(self.compressed_anim('cy', y_path, cfg.time_scale))
                target.add_animation(self.compressed_anim('cx', x_path, cfg.time_scale))
            target.add_animation(self.compressed_anim("visibility", visibility, cfg.time_scale))

    def create_obstacles(self, grid_holder, animation_config):
        """
        Creates the obstacles.
        :param grid_holder:
        :param animation_config:
        :return:
        """
        gh = grid_holder
        cfg = self.svg_settings

        result = []

        for i in range(gh.height):
            for j in range(gh.width):
                x, y = self.fix_point(i, j, gh.width)

                if gh.obstacles[x][y] != self.grid_config.FREE:
                    obs_settings = {}
                    obs_settings.update(x=cfg.draw_start + i * cfg.scale_size - cfg.r,
                                        y=cfg.draw_start + j * cfg.scale_size - cfg.r,
                                        height=cfg.r * 2,
                                        )

                    if animation_config.egocentric_idx is not None and cfg.egocentric_shaded:
                        initial_positions = [agent_states[0].get_xy() for agent_states in gh.history]
                        ego_x, ego_y = initial_positions[animation_config.egocentric_idx]
                        if not self.check_in_radius(x, y, ego_x, ego_y, self.grid_config.obs_radius):
                            obs_settings.update(opacity=cfg.shaded_opacity)

                    result.append(RectangleHref(**obs_settings))

        return result

    def animate_obstacles(self, obstacles, grid_holder, animation_config):
        """

        :param obstacles:
        :param grid_holder:
        :param animation_config:
        :return:
        """
        gh: GridHolder = grid_holder
        obstacle_idx = 0
        cfg = self.svg_settings

        for i in range(gh.height):
            for j in range(gh.width):
                x, y = self.fix_point(i, j, gh.width)
                if gh.obstacles[x][y] == self.grid_config.FREE:
                    continue
                opacity = []
                seen = set()
                for step_idx, agent_state in enumerate(gh.history[animation_config.egocentric_idx]):
                    ego_x, ego_y = agent_state.get_xy()
                    if self.check_in_radius(x, y, ego_x, ego_y, self.grid_config.obs_radius):
                        seen.add((x, y))
                    if (x, y) in seen:
                        opacity.append(str(1.0))
                    else:
                        opacity.append(str(cfg.shaded_opacity))

                obstacle = obstacles[obstacle_idx]
                obstacle.add_animation(self.compressed_anim('opacity', opacity, cfg.time_scale))

                obstacle_idx += 1

    def create_agents(self, grid_holder, animation_config):
        initial_positions = [state[0].get_xy() for state in grid_holder.history if state[0].is_active()]
        ego_idx = animation_config.egocentric_idx
        agents = []

        for idx, (x, y) in enumerate(initial_positions):
            circle_settings = {
                'cx': self.svg_settings.draw_start + y * self.svg_settings.scale_size,
                'cy': self.svg_settings.draw_start + (grid_holder.width - x - 1) * self.svg_settings.scale_size,
                # 'r': self.svg_settings.r,
                'fill': grid_holder.colors[idx],
                'class': 'agent',
            }

            if ego_idx is not None:
                ego_x, ego_y = initial_positions[ego_idx]
                is_out_of_radius = not self.check_in_radius(x, y, ego_x, ego_y, self.grid_config.obs_radius)
                circle_settings['fill'] = self.svg_settings.ego_other_color
                if idx == ego_idx:
                    circle_settings['fill'] = self.svg_settings.ego_color
                elif is_out_of_radius and self.svg_settings.egocentric_shaded:
                    circle_settings['opacity'] = self.svg_settings.shaded_opacity

            agents.append(Circle(**circle_settings))

        return agents

    def create_targets(self, grid_holder, animation_config):
        """
        Creates the targets.
        :param grid_holder:
        :param animation_config:
        :return:
        """
        gh: GridHolder = grid_holder
        cfg = self.svg_settings
        targets = []
        for agent_idx, agent_states in enumerate(gh.history):

            tx, ty = agent_states[0].get_target_xy()
            x, y = ty, gh.width - tx - 1

            if not any([agent_state.is_active() for agent_state in gh.history[agent_idx]]):
                continue

            circle_settings = {"class": 'target'}
            circle_settings.update(cx=cfg.draw_start + x * cfg.scale_size,
                                   cy=cfg.draw_start + y * cfg.scale_size,
                                   # r=cfg.r,
                                   stroke=gh.colors[agent_idx],
                                   # stroke_width=cfg.stroke_width,
                                   # fill='none'
                                   )
            if animation_config.egocentric_idx is not None:
                if animation_config.egocentric_idx != agent_idx:
                    continue

                circle_settings.update(stroke=cfg.ego_color)
            target = Circle(**circle_settings)
            targets.append(target)
        return targets


def main():
    grid = """
    ...#.
    .#...
    .....
    ##.#.
    """
    grid_config = GridConfig(size=32, num_agents=2, obs_radius=2, seed=9, on_target='restart', max_episode_steps=16,
                             density=0.1, map=grid, observation_type="POMAPF")
    env = pogema_v0(grid_config=grid_config)
    env = AnimationMonitor(env)

    obs, _ = env.reset()
    truncated = terminated = [False]

    agent = BatchAStarAgent()
    while not all(terminated) and not all(truncated):
        obs, _, terminated, truncated, _ = env.step(agent.act(obs))

    env.save_animation('out-static.svg', AnimationConfig(static=True, save_every_idx_episode=None))
    env.save_animation('out-static-ego.svg', AnimationConfig(egocentric_idx=0, static=True))
    env.save_animation('out-static-no-agents.svg', AnimationConfig(show_agents=False, static=True))
    env.save_animation("out.svg")
    env.save_animation("out-ego.svg", AnimationConfig(egocentric_idx=0))


if __name__ == '__main__':
    main()



class SvgObject:
    tag = None

    def __init__(self, **kwargs):
        self.attributes = kwargs
        self.animations = []

    def add_animation(self, animation):
        self.animations.append(animation)

    @staticmethod
    def render_attributes(attributes):
        result = " ".join([f'{x.replace("_", "-")}="{y}"' for x, y in sorted(attributes.items())])
        return result

    def render(self):
        animations = '\n'.join([a.render() for a in self.animations]) if self.animations else None
        if animations:
            return f"<{self.tag} {self.render_attributes(self.attributes)}> {animations} </{self.tag}>"
        return f"<{self.tag} {self.render_attributes(self.attributes)} />"


class Rectangle(SvgObject):
    """
    Rectangle class for the SVG.
    """
    tag = 'rect'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.attributes['y'] = -self.attributes['y'] - self.attributes['height']


class RectangleHref(SvgObject):
    """
    Rectangle class for the SVG.
    """
    tag = 'use'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.attributes['y'] = -self.attributes['y'] - self.attributes['height']
        self.attributes['href'] = "#obstacle"
        del self.attributes['height']


class Circle(SvgObject):
    """
    Circle class for the SVG.
    """
    tag = 'circle'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.attributes['cy'] = -self.attributes['cy']


class Line(SvgObject):
    """
    Line class for the SVG.
    """
    tag = 'line'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.attributes['y1'] = -self.attributes['y1']
        self.attributes['y2'] = -self.attributes['y2']


class Animation(SvgObject):
    """
    Animation class for the SVG.
    """
    tag = 'animate'

    def render(self):
        return f"<{self.tag} {self.render_attributes(self.attributes)}/>"



import sys

from pydantic import BaseModel

from typing_extensions import Literal


class AgentsTargetsSizeError(Exception):
    pass


def check_grid(obstacles, agents_xy, targets_xy):
    if bool(agents_xy) != bool(targets_xy):
        raise AgentsTargetsSizeError("Agents and targets must be defined together/undefined together!")

    if not agents_xy or not targets_xy:
        return

    if len(agents_xy) != len(targets_xy):
        raise IndexError("Can't create task. Please provide agents_xy and targets_xy of the same size.")

    # check overlapping of agents
    for i in range(len(agents_xy)):
        for j in range(i + 1, len(agents_xy)):
            if agents_xy[i] == agents_xy[j]:
                raise ValueError(f"Agents can't overlap! {agents_xy[i]} is in both {i} and {j} position.")

    for start_xy, finish_xy in zip(agents_xy, targets_xy):
        s_x, s_y = start_xy
        if obstacles[s_x, s_y]:
            raise KeyError(f'Cell is {s_x, s_y} occupied by obstacle.')
        f_x, f_y = finish_xy
        if obstacles[f_x, f_y]:
            raise KeyError(f'Cell is {f_x, f_y} occupied by obstacle.')

    # todo check connectivity of starts and finishes


def render_grid(obstacles, positions_xy=None, targets_xy=None, is_active=None, mode='human'):
    if positions_xy is None:
        positions_xy = []
    if targets_xy is None:
        targets_xy = []
    if is_active is None:
        if positions_xy:
            is_active = [True] * len(positions_xy)
        else:
            is_active = []
    from io import StringIO
    import string
    from gymnasium import utils as gym_utils
    from contextlib import closing

    outfile = StringIO() if mode == 'ansi' else sys.stdout
    chars = string.digits + string.ascii_letters + string.punctuation
    positions_map = {(x, y): id_ for id_, (x, y) in enumerate(positions_xy) if is_active[id_]}
    finishes_map = {(x, y): id_ for id_, (x, y) in enumerate(targets_xy) if is_active[id_]}
    for line_index, line in enumerate(obstacles):
        out = ''
        for cell_index, cell in enumerate(line):
            if cell == CommonSettings().FREE:
                agent_id = positions_map.get((line_index, cell_index), None)
                finish_id = finishes_map.get((line_index, cell_index), None)

                if agent_id is not None:
                    out += str(gym_utils.colorize(' ' + chars[agent_id % len(chars)] + ' ', color='red', bold=True,
                                                  highlight=False))
                elif finish_id is not None:
                    out += str(
                        gym_utils.colorize('|' + chars[finish_id % len(chars)] + '|', 'white', highlight=False))
                else:
                    out += str(gym_utils.colorize(str(' . '), 'white', highlight=False))
            else:
                out += str(gym_utils.colorize(str('   '), 'cyan', bold=False, highlight=True))
        out += '\n'
        outfile.write(out)

    if mode != 'human':
        with closing(outfile):
            return outfile.getvalue()


class CommonSettings(BaseModel):
    MOVES: list = [[0, 0], [-1, 0], [1, 0], [0, -1], [0, 1], ]
    FREE: Literal[0] = 0
    OBSTACLE: Literal[1] = 1
    empty_outside: bool = True
